@@ -1,8 +1,7 @@
 import io
-from random import shuffle
+from random import shuffle, choice
 from json import loads, dumps
 from pathlib import Path
-from secrets import token_urlsafe
 from typing import List, Dict, Optional
 import re
 import textwrap
@@ -11,8 +10,11 @@ from requests import get, head
 from requests.exceptions import ConnectionError, MissingSchema, TooManyRedirects, ChunkedEncodingError, ReadTimeout
 from PIL import Image, ImageFont, ImageDraw, UnidentifiedImageError, ImageOps
 from PIL.PngImagePlugin import PngImageFile
+from fontTools.ttLib import TTFont
+from unidecode import unidecode
+import markovify
 from perlin_numpy.perlin2d import generate_fractal_noise_2d
-from procemon.paths import TYPES_DIRECTORY, IMAGES_DIRECTORY, FONTS_DIRECTORY, MOVES_DIRECTORY
+from procemon.paths import TYPES_DIRECTORY, IMAGES_DIRECTORY, TEXT_FONT, SYMBOL_FONT, MOVES_DIRECTORY, REGIONS_DIRECTORY
 from procemon.monster_type import MonsterType
 from procemon.monster import Monster
 from procemon.rarity import Rarity
@@ -25,26 +27,6 @@ class Dex:
     When created, the dex will randomly select types and moods, assign verbs and adjectives, and generate the Procemon.
     """
 
-    """:class_var
-    The path to the file of URLs that we know are bad.
-    """
-    BAD_IMAGE_URLS_PATH: Path = IMAGES_DIRECTORY.joinpath("bad_image_urls.txt")
-    """:class_var
-    A list of known bad URLs.
-    """
-    BAD_IMAGE_URLS: List[str] = BAD_IMAGE_URLS_PATH.read_text(encoding="utf-8").split("\n")
-    """:class_var
-    The path to the file of URLs that we know are bad.
-    """
-    BAD_WNIDS_PATH: Path = IMAGES_DIRECTORY.joinpath("bad_wnids.txt")
-    """:class_var
-    A list of known bad wnids.
-    """
-    BAD_WNIDS: List[str] = BAD_WNIDS_PATH.read_text(encoding="utf-8").split("\n")
-    """:class_var
-    Imagenet data. Key = A word. Value = The wnid corresponding to that word.
-    """
-    IMAGENET: Dict[str, str] = loads(IMAGES_DIRECTORY.joinpath("imagenet_wnids.json").read_text(encoding="utf-8"))
     """:class_var
     A numpy array of a color palette.
     """
@@ -65,6 +47,20 @@ class Dex:
     The path to the energy icons. 
     """
     ENERGY_DIRECTORY: Path = IMAGES_DIRECTORY.joinpath("energy")
+    with TTFont(TEXT_FONT) as font:
+        """:class_var
+        A list of all Unicode characters supported by the font. Source: https://stackoverflow.com/a/58232763
+        """
+        SUPPORTED_CHARACTERS: List[str] = list(set(chr(y[0]) for x in font["cmap"].tables for y in x.cmap.items()))
+    """:class_var
+    Base URL for the Wikipedia API. Source: https://stackoverflow.com/a/41807620
+    """
+    WIKIPEDIA_API_URL: str = "https://en.wikipedia.org/w/api.php?action=query&prop=imageinfo&format=json&iiprop=url" \
+                             "&generator=images&titles="
+    """:class_var
+    Ignore these image URLs.
+    """
+    URL_EXCLUDE: List[str] = ["https://upload.wikimedia.org/wikipedia/commons/7/74/Red_Pencil_Icon.png"]
 
     def __init__(self, num_types: int = 12, num_monsters_per_type: int = 9, quiet: bool = False):
         """
@@ -74,10 +70,7 @@ class Dex:
         """
 
         # Get all of the types.
-        all_types: List[MonsterType] = list()
-        for f in TYPES_DIRECTORY.iterdir():
-            td = loads(f.read_text(encoding="utf-8"))
-            all_types.append(MonsterType(**td))
+        all_types = Dex.get_all_types()
         # Get a random subset of the types.
         shuffle(all_types)
         all_types = all_types[:num_types]
@@ -87,14 +80,35 @@ class Dex:
         self.types: Dict[str, MonsterType] = dict()
         for t in all_types:
             self.types[t.monster_type] = t
-
+        color_indices: List[int] = list(np.arange(len(Dex.LIGHT_COLORS)))
+        shuffle(color_indices)
+        """:field
+        The indices of colors in the palette mapped to names of monster types.
+        """
+        self.color_indices: Dict[str, int] = dict()
+        color_index: int = 0
+        for t in all_types:
+            self.color_indices[t.monster_type] = color_indices[color_index]
+            color_index += 1
+            # If there are more monster types than colors, go back to the start of the color index list.
+            if color_index >= len(color_indices):
+                color_index = 0
         attack_verbs = MOVES_DIRECTORY.joinpath("attack_verbs.txt").read_text(encoding="utf-8").split("\n")
         shuffle(attack_verbs)
 
+        """:field
+        The name of the region of the dex.
+        """
+        self.region: str = Dex.get_region()
+        """:field
+        A random dingbat for the region.
+        """
+        self.region_symbol: str = Dex.get_region_symbol()
+
         """:field
         The output directory of the dex.
         """
-        self.dst: Path = Path(f"dst/dex/{token_urlsafe(3)}")
+        self.dst: Path = Path(f"dst/dex/{self.region}")
         if not self.dst.exists():
             self.dst.mkdir(parents=True)
         if not quiet:
@@ -147,25 +161,19 @@ class Dex:
         """
         self.images_per_type: Dict[str, List[PngImageFile]] = dict()
 
-        color_indices: List[int] = list(np.arange(len(all_types)))
-        shuffle(color_indices)
-        """:field
-        The indices of colors in the palette mapped to names of monster types.
-        """
-        self.color_indices: Dict[str, int] = dict()
-        for t, i in zip(all_types, color_indices):
-            self.color_indices[t.monster_type] = i
-
     def write_json(self) -> None:
         """
         Save the dex as a JSON dictionary.
         """
 
-        data = dict()
+        monsters = dict()
         for t in self.monsters:
-            data[t] = dict()
+            monsters[t] = dict()
             for n in self.monsters[t]:
-                data[t][n] = self.monsters[t][n].__dict__
+                monsters[t][n] = self.monsters[t][n].__dict__
+        data = {"region": self.region,
+                "symbol": self.region_symbol,
+                "dex": monsters}
         self.dst.joinpath("dex.json").write_text(dumps(data, sort_keys=True, indent=2, cls=DexEncoder),
                                                  encoding="utf-8")
 
@@ -217,54 +225,89 @@ class Dex:
         :return: A list of converted images for this type using ImageNet data.
         """
 
-        # Get the wnids of the type name and the nouns. Some words might not have wnids.
-        wnids = []
-        for n in self.types[monster_type].nouns:
-            if n in Dex.IMAGENET:
-                wnids.append(Dex.IMAGENET[n])
-        # Randomize the order of the wnids.
-        shuffle(wnids)
-        # Insert the root wnid at start.
-        wnids.insert(0, Dex.IMAGENET[self.types[monster_type].imagenet])
+        # Get a list of nouns that are known not to have any images.
+        no_images_path = IMAGES_DIRECTORY.joinpath("no_images.txt")
+        no_images = no_images_path.read_text(encoding="utf-8").split("\n")
 
-        images: List[PngImageFile] = list()
+        # Get the words used for each monster in the dex.
+        words: List[str] = [m.words[0] for m in self.monsters[monster_type].values()]
 
-        # Get a list of URLs from the list of wnids.
-        wnid_index = 0
-        while wnid_index < len(wnids) and len(images) < self.__num_monsters_per_type:
-            wnid = wnids[wnid_index]
-            # Get URLs from the wnid.
-            urls = Dex.get_wnid_urls(wnid=wnid)
-            if len(urls) == 0:
-                wnid_index += 1
-                if wnid not in Dex.BAD_WNIDS:
-                    Dex.add_to_bad_urls(wnid)
-                continue
-            # We'll only test these and move on if they're no good.
-            urls = urls[:20]
-            # Test each URL.
-            for url in urls:
-                img = Dex.get_image_from_url(url=url)
-                # Remember that this is a bad image so we never try it again.
-                if img is None:
-                    Dex.add_to_bad_urls(url)
+        # Use these nouns if the noun in the list fails.
+        fallback_nouns = self.types[monster_type].nouns[:]
+        # Skip any nouns that are known to not have images.
+        fallback_nouns = [n for n in fallback_nouns if n not in no_images]
+        # Randomize the list of nouns.
+        shuffle(fallback_nouns)
+        # A dictionary of images, where the key is the URL.
+        images: Dict[str, PngImageFile] = dict()
+
+        for i in range(len(words)):
+            # Prefer a Wikipedia page with the same name as the word.
+            nouns = [words[i]]
+            # Fallback: Any other noun in the monster type.
+            nouns.extend(fallback_nouns)
+            # Fallback: The general Wikipedia page.
+            nouns.append(self.types[monster_type].wikipedia)
+            got_image = False
+            for n in nouns:
+                if got_image:
+                    break
+                # Try to get images from a Wikipedia page. If the page doesn't exist, remember not to try it again.
+                try:
+                    resp = get(f"{Dex.WIKIPEDIA_API_URL}{n}", timeout=20)
+                    if resp.status_code != 200 and resp.status_code != 301:
+                        no_images.append(n)
+                        continue
+                    data = resp.json()
+                except ConnectionError:
+                    no_images.append(n)
                     continue
-                # Convert to grayscale.
-                img = ImageOps.grayscale(img)
-                # Increase the contrast.
-                img = ImageOps.autocontrast(img)
-                # Resize.
-                img = img.resize((32, 32), Image.LANCZOS)
-                # Colorize using the palette color for this type.
-                img = ImageOps.colorize(img, black="black",
-                                        white=Dex.LIGHT_COLORS[self.color_indices[monster_type]])
-                # Enlarge.
-                img = img.resize((400, 400), Image.NEAREST)
-                # Append the image.
-                images.append(img)
-            wnid_index += 1
-        shuffle(images)
-        return images
+                except ReadTimeout:
+                    no_images.append(n)
+                    continue
+                # This page doesn't exist.
+                if "query" not in data:
+                    no_images.append(n)
+                    continue
+                urls: List[str] = list()
+                for page in data["query"]["pages"]:
+                    # Pages with IDs are logos or icons.
+                    if "pageid" in data["query"]["pages"][page]:
+                        continue
+                    for image_info in data["query"]["pages"][page]["imageinfo"]:
+                        urls.append(image_info["url"])
+                if len(urls) == 0:
+                    no_images.append(n)
+                    continue
+                for url in urls:
+                    # Skip URLs that we've already added.
+                    # Skip svg files because they're usually maps, icons, logos, etc.
+                    # Skip any other known logos.
+                    if url in images or ".svg" in url or url.endswith(".webm") or url in Dex.URL_EXCLUDE:
+                        continue
+                    # Try to convert the image URL into a PIL image.
+                    img = Dex.get_image_from_url(url=url)
+                    if img is None:
+                        continue
+                    # Convert to grayscale.
+                    img = ImageOps.grayscale(img)
+                    # Increase the contrast.
+                    img = ImageOps.autocontrast(img)
+                    # Resize.
+                    img = img.resize((32, 32), Image.LANCZOS)
+                    # Colorize using the palette color for this type.
+                    img = ImageOps.colorize(img, black="black",
+                                            white=Dex.LIGHT_COLORS[self.color_indices[monster_type]])
+                    # Enlarge.
+                    img = img.resize((400, 400), Image.NEAREST)
+                    # Append the image.
+                    images[url] = img
+                    # Got an image for this card.
+                    got_image = True
+                    break
+        # Remember the nouns that don't have images.
+        no_images_path.write_text(("\n".join(list(sorted(set(no_images))))).strip(), encoding="utf-8")
+        return list(images.values())
 
     def get_card(self, monster: Monster) -> PngImageFile:
         """
@@ -273,6 +316,13 @@ class Dex:
         :return: A card image for this monster.
         """
 
+        # Make sure that the monster's description string is supported by the card font.
+        # We only check the description because we know that all names, types, verbs, and adjectives are ok.
+        # See: `util/font_test.py` in the repo.
+        if monster.description is None:
+            monster.description = "None"
+        monster.description = Dex.get_supported_string(monster.description)
+
         card = Image.open(str(Dex.CARD_PATH.resolve()))
 
         # Get Perlin noise.
@@ -290,12 +340,8 @@ class Dex:
                 if pixels[x, y] == (255, 255, 255, 255):
                     # Add some Perlin noise.
                     pixels[x, y] = Dex.lighten(bg_color, (perlin_noise[y, x]))
-
-        # Get the font file.
-        font_file: str = str(FONTS_DIRECTORY.joinpath("pokemon-classic.ttf").resolve())
-
         pad_x = 52
-
+        font_file = str(TEXT_FONT.resolve())
         # Add the name of the monster.
         f_header = ImageFont.truetype(font_file, 28)
         draw = ImageDraw.Draw(card)
@@ -419,8 +465,9 @@ class Dex:
                 line_size = f_move_special.getsize(line)
                 line_y += line_size[1] + 8
             # If there's too much special text, don't print it!
-            if line_y >= 940:
+            if move_text_y + move_font_text_size[1] + 12 >= 930:
                 m.special = ""
+                lines.clear()
                 m.damage = 1
 
             if m.damage > 0:
@@ -505,30 +552,6 @@ class Dex:
 
         return card
 
-    @staticmethod
-    def add_to_bad_urls(url: str) -> None:
-        """
-        Remember that this a bad URL.
-
-        :param url: The bad URL.
-        """
-
-        Dex.BAD_IMAGE_URLS.append(url)
-        with io.open(str(Dex.BAD_IMAGE_URLS_PATH.resolve()), "at", encoding="utf-8") as f:
-            f.write(url + "\n")
-
-    @staticmethod
-    def add_to_bad_wnids(wnid: str) -> None:
-        """
-        Remember that this a bad wnid.
-
-        :param wnid: The wnid URL.
-        """
-
-        Dex.BAD_WNIDS.append(wnid)
-        with io.open(str(Dex.BAD_WNIDS_PATH.resolve()), "at", encoding="utf-8") as f:
-            f.write(wnid + "\n")
-
     @staticmethod
     def lighten(color, percent) -> tuple:
         """
@@ -576,6 +599,8 @@ class Dex:
         # If it takes too long to get the image, assume that it doesn't exist.
         except ReadTimeout:
             return None
+        except MissingSchema:
+            return None
 
         # Try to get the image.
         try:
@@ -603,40 +628,68 @@ class Dex:
             return None
 
     @staticmethod
-    def get_wnid_urls(wnid: str) -> List[str]:
+    def get_supported_string(string: str) -> str:
         """
-        Try to get image URLs from a wnid.
+        :param string: A string that might have characters that the card font doesn't support.
 
-        :param wnid: The wnid.
-
-        :return: A list of image URLs in the wnid. Can be empty.
+        :return: A converted string in which all characters are supported by the card font.
         """
-        if wnid.startswith("-"):
-            wnid = wnid[1:]
-        if wnid in Dex.BAD_WNIDS:
-            return []
-        got_resp = False
-        resp = None
-        num_attempts = 0
-        while not got_resp and num_attempts < 10:
-            # Try to get the data. We know all of these URLs are valid.
-            try:
-                resp = get(f"http://www.image-net.org/api/text/imagenet.synset.geturls.getmapping?wnid={wnid}")
-                got_resp = True
-            except ConnectionError:
-                num_attempts += 1
-                continue
-        # If this wnid is totally bogus, remember not to try it again.
-        if not got_resp or "Invalid" in resp.content.decode("utf-8"):
-            Dex.add_to_bad_wnids(wnid)
-            return []
-        # Get all of the image URLs.
-        urls = resp.content.decode("utf-8").split("\r\n")
-        urls = [url.split(" ")[1].strip() for url in urls if len(url.strip()) > 0 and
-                url.split(" ")[1].strip() not in Dex.BAD_IMAGE_URLS]
-        # If there are no URLs, this is a bad wnid.
-        if len(urls) == 0:
-            Dex.add_to_bad_wnids(wnid)
-        # Sometimes there's just whole wnids of bad URLs. We don't need to test them all. We've got places to be!
-        shuffle(urls)
-        return urls
+
+        return "".join([(c if c in Dex.SUPPORTED_CHARACTERS else unidecode(c)) for c in string])
+
+    @staticmethod
+    def get_all_types() -> List[MonsterType]:
+        """
+        :return: A list of all available `MonsterTypes`.
+        """
+
+        all_types: List[MonsterType] = list()
+        for f in TYPES_DIRECTORY.iterdir():
+            if f.is_file() and f.suffix == ".json":
+                td = loads(f.read_text(encoding="utf-8"))
+                all_types.append(MonsterType(**td))
+        return all_types
+
+    @staticmethod
+    def get_region() -> str:
+        """
+        :return: The name of the region of the dex.
+        """
+
+        japan = REGIONS_DIRECTORY.joinpath("japan.txt").read_text(encoding="utf-8").split("\n")
+        # Parse Japanese placenames into syllables and treat those as words.
+        placenames = list()
+        for j in japan:
+            syllables = re.findall(r"([b-df-hj-np-tv-z]+[aeiouyūō]|[aeiouyūō][b-df-hj-np-tv-yz]+|[aeiouūō])",
+                                   Dex.get_supported_string(j.lower()))
+            syllables[0] = syllables[0].title()
+            sentence = (" ".join(syllables) + ".").strip()
+            placenames.append(sentence)
+        # Get a list of US county names for spice. Take only a subset of them. Treat each letter as a word.
+        us = REGIONS_DIRECTORY.joinpath("us.txt").read_text(encoding="utf-8").split("\n")
+        shuffle(us)
+        placenames.extend([(" ".join(list(u)) + ".").strip() for u in us[:100]])
+        # Create a markov model.
+        model = markovify.Text(placenames)
+        # Create a "sentence" and convert it to a word.
+        try:
+            region = model.make_sentence(tries=100, max_words=12).replace(" ", "")[:-1].lower().title()
+            # Add a vowel to the end of the name.
+            if region[-1] == "e" or region[-1] == "s":
+                return region
+            if region[-1] not in "aiou":
+                region += choice("aiou")
+            return region
+        # Fallback if we fail to make the name.
+        except AttributeError:
+            return "Mystery"
+
+    @staticmethod
+    def get_region_symbol() -> str:
+        """
+        :return: A symbol for the region.
+        """
+
+        with TTFont(str(SYMBOL_FONT.resolve())) as font:
+            chars = list(set(chr(y[0]) for x in font["cmap"].tables for y in x.cmap.items() if chr(y[0]).isalnum()))
+        return choice(chars)
